import java.util.*;
import java.io.*;

public class RomanNumeral implements Serializable
{

	String name; // Roman Numeral
	int value; // integer value

	// Hashtable to keep up with existing Roman Numerals
	private static Hashtable ht = new Hashtable();

	// base Roman Numerals
	public static final RomanNumeral I = RomanNumeral.create("I", 1);
	public static final RomanNumeral V = RomanNumeral.create("V", 5);
	public static final RomanNumeral X = RomanNumeral.create("X", 10);
	public static final RomanNumeral L = RomanNumeral.create("L", 50);
	public static final RomanNumeral C = RomanNumeral.create("C", 100);
	public static final RomanNumeral D = RomanNumeral.create("D", 500);
	public static final RomanNumeral M = RomanNumeral.create("M", 1000);

	// creates RomanNumeral from String
	public static RomanNumeral create(String n)
	{
		return create(n, toInt(n));
	}

	// generates our RomanNumerals guaranteeing that they are unique
	// (no validation so IIIII will be treated as 5)
	private static RomanNumeral create(String n, int i)
	{
		if(ht.containsKey(n))
		{
			return (RomanNumeral)(ht.get(n));
		}
		else
		{
			RomanNumeral rn = new RomanNumeral(n, i);
			ht.put(n, rn);
			return rn;
		}
	}

	// private constructor called from create()
	private RomanNumeral(String n, int i)
	{
		name = n;
		value = i;
	}

	// called just before the object is deserialized; the object
	// returned here is what actually will be deserialized
	public Object readResolve() throws InvalidObjectException
	{

		if(!ht.containsKey(name))
		{
			ht.put(name, this);
		}

		RomanNumeral instance = (RomanNumeral) ht.get(name);
		if (instance != null) 
		{
			return instance;
		} 
		else 
		{
			throw new InvalidObjectException("Unknown Roman Numeral");
		}

	}

	public String toString()
	{
		return name;
	}

	// converts this Roman Numeral to int
	public int toInt()
	{
		return toInt(name);
	}

	// converts any Roman Numeral to int (no validation)
	public static int toInt(String inName)
	{
		StringBuffer sb = new StringBuffer(inName);
		int retInt = 0;
		
		// for each letter in the Roman Numeral
		for(int i=0; i<sb.length(); i++)
		{
			// get int value
			int rnVal1 = rnCharToInt(sb.charAt(i));

			// check if not last letter
			if(i < sb.length()-1)
			{
				// get next value and compare to this one
				int rnVal2 = rnCharToInt(sb.charAt(i+1));

				if(rnVal1 < rnVal2) // if this one less than next
				{
					// need to subtract and skip next
					retInt += (rnVal2 - rnVal1);
					i++;
				}
				else // add as normal
				{
					retInt += rnVal1;
				}

			}
			else // add as normal
			{
				retInt += rnVal1;
			}
		}

		return retInt;
	}

	// returns the int value of a char that represents I,V,X,L,C,D or,M
	private static int rnCharToInt(char c)
	{
		String inVal = String.valueOf(c);
		return ((RomanNumeral)(ht.get(inVal))).value;
	}

	// creates RomanNumeral from int
	// logic used here is from Javascript code written by David Turley
	public static RomanNumeral create(int inVal)
	{
		int num = inVal;

		String onesArray[] = {"I","II","III","IV","V","VI","VII","VIII","IX"};
		String tensArray[] = {"X","XX","XXX","XL","L","LX","LXX","LXXX","XC"};
		String hundredsArray[] = {"C","CC","CCC","CD","D","DC","DCC","DCCC","CM"};

		int ones = num % 10;
		num = (num - ones) / 10;
		int tens = num % 10;
		num = (num - tens) / 10;
		int hundreds = num % 10;
		num = (num - hundreds) / 10;

		String Roman = "";
    
		for (int i=0; i < num; i++)
		{
			Roman += "M";
		}

		if (hundreds > 0) 
		{
			Roman += hundredsArray[hundreds-1];
		}
    
		if (tens > 0) 
		{
			Roman += tensArray[tens-1];
		}
    
		if (ones > 0) 
		{
			Roman += onesArray[ones-1];
		}

		return create(Roman);
	}

	// I'm not sure why someone would want to do arithmetic
	// on roman numerals but I was on a roll

	public RomanNumeral plus(RomanNumeral rnIn)
	{
		return create(this.toInt() + rnIn.toInt());
	}

	public RomanNumeral minus(RomanNumeral rnIn)
	{
		return create(this.toInt() - rnIn.toInt());
	}

	public RomanNumeral times(RomanNumeral rnIn)
	{
		return create(this.toInt() * rnIn.toInt());
	}

	public RomanNumeral dividedBy(RomanNumeral rnIn)
	{
		return create(this.toInt() / rnIn.toInt());
	}

	public RomanNumeral mod(RomanNumeral rnIn)
	{
		return create(this.toInt() % rnIn.toInt());
	}

	public boolean isLessThan(RomanNumeral rnIn)
	{
		return (this.toInt() < rnIn.toInt());
	}

	public boolean isGreaterThan(RomanNumeral rnIn)
	{
		return (this.toInt() > rnIn.toInt());
	}

}